@@ -32,6 +32,7 @@
 from ctypes import *
 from ctypes.util import find_library
 import sys
+import threading
     
 ## C Type mappings ##
 ## Enums
@@ -99,20 +100,10 @@ NVML_ERROR_ALREADY_INITIALIZED = 5
 NVML_ERROR_NOT_FOUND           = 6
 NVML_ERROR_INSUFFICIENT_SIZE   = 7
 NVML_ERROR_INSUFFICIENT_POWER  = 8
+NVML_ERROR_DRIVER_NOT_LOADED   = 9
+NVML_ERROR_TIMEOUT             = 10,
 NVML_ERROR_UNKNOWN             = 999
 
-_compat2_0ErrorToString = {
-    NVML_SUCCESS                   : 'Success',
-    NVML_ERROR_UNINITIALIZED       : 'Uninitialized',
-    NVML_ERROR_INVALID_ARGUMENT    : 'Invalid Argument',
-    NVML_ERROR_NOT_SUPPORTED       : 'N/A',
-    NVML_ERROR_NO_PERMISSION       : 'Insufficient Permissions',
-    NVML_ERROR_ALREADY_INITIALIZED : 'Already Initialized',
-    NVML_ERROR_NOT_FOUND           : 'Not Found',
-    NVML_ERROR_INSUFFICIENT_SIZE   : 'Insufficient Size',
-    NVML_ERROR_UNKNOWN             : 'Unknown Error',
-}
-
 _nvmlFanState_t = c_uint
 NVML_FAN_NORMAL             = 0
 NVML_FAN_FAILED             = 1
@@ -125,28 +116,18 @@ NVML_LED_COLOR_AMBER        = 1
 nvmlFlagDefault             = 0
 nvmlFlagForce               = 1
 
+NVML_VALUE_NOT_AVAILABLE_ulonglong = c_ulonglong(-1)
+
 ## Lib loading ##
-nvml = None
-try:
-    if (sys.platform[:3] == "win"):
-        # cdecl calling convention
-        nvml = cdll.nvml
-    else:
-        # assume linux
-        nvml = CDLL("libnvidia-ml.so")
-except OSError, ose:
-    print ose
-    sys.exit(ose.errno)
-if (nvml == None):
-    print "Failed to load NVML"
-    sys.exit(1)
+nvmlLib = None
+libLoadLock = threading.Lock()
 
 ## Error Checking ##
 class NVMLError(Exception):
     def __init__(self, value):
         self.value = value
     def __str__(self):
-        return nvmlErrorString(self.value)
+        return str(nvmlErrorString(self.value))
 
 def _nvmlCheckReturn(ret):
     if (ret != NVML_SUCCESS):
@@ -155,11 +136,49 @@ def _nvmlCheckReturn(ret):
 
 ## Function access ##
 def _nvmlGetFunctionPointer(name):
+    global nvmlLib
+    global libLoadLock
+    
+    libLoadLock.acquire()
     try:
-        fn = getattr(nvml, name)
-    except AttributeError, attrError:
-        raise NVMLError(NVML_ERROR_NOT_SUPPORTED)
-    return fn
+        # ensure library was loaded
+        if (nvmlLib == None):
+            raise NVMLError(NVML_ERROR_UNINITIALIZED)
+        try:
+            return getattr(nvmlLib, name)
+        except AttributeError, attrError:
+            raise NVMLError(NVML_ERROR_NOT_SUPPORTED)
+    finally:
+        # lock is always freed
+        libLoadLock.release()
+
+## Alternative object
+# Allows the object to be printed
+# Allows mismatched types to be assigned
+#  - like None when the Structure variant requires c_uint
+class nvmlFriendlyObject(object):
+    def __init__(self, dictionary):
+        for x in dictionary:
+            setattr(self, x, dictionary[x])
+    def __str__(self):
+        return self.__dict__.__str__()
+
+def nvmlStructToFriendlyObject(struct):
+    d = {}
+    for x in struct._fields_:
+        key = x[0]
+        value = getattr(struct, key)
+        d[key] = value
+    obj = nvmlFriendlyObject(d)
+    return obj
+
+# pack the object so it can be passed to the NVML library
+def nvmlFriendlyObjectToStruct(obj, model):
+    for x in model._fields_:
+        key = x[0]
+        value = obj.__dict__[key]
+        setattr(model, key, value)
+    return model
 
 ## Unit structures
 class struct_c_nvmlUnit_t(Structure):
@@ -212,6 +231,13 @@ class nvmlPciInfo_t(Structure):
         ('bus', c_uint),
         ('device', c_uint),
         ('pciDeviceId', c_uint),
+        
+        # Added in 2.285
+        ('pciSubSystemId', c_uint),
+        ('reserved0', c_uint),
+        ('reserved1', c_uint),
+        ('reserved2', c_uint),
+        ('reserved3', c_uint),
     ]
 
 class c_nvmlMemory_t(Structure):
@@ -221,6 +247,22 @@ class c_nvmlMemory_t(Structure):
         ('used', c_ulonglong),
     ]
 
+# On Windows with the WDDM driver, usedGpuMemory is reported as None
+# Code that processes this structure should check for None, I.E.
+#
+# if (info.usedGpuMemory == None):
+#     # TODO handle the error
+#     pass
+# else:
+#    print("Using %d MB of memory" % (info.usedGpuMemory / 1024 / 1024))
+#
+# See NVML documentation for more information
+class c_nvmlProcessInfo_t(Structure):
+    _fields_ = [
+        ('pid', c_uint),
+        ('usedGpuMemory', c_ulonglong),
+    ]
+
 class c_nvmlEccErrorCounts_t(Structure):
     _fields_ = [
         ('l1Cache', c_ulonglong),
@@ -235,34 +277,110 @@ class c_nvmlUtilization_t(Structure):
         ('memory', c_uint),
     ]
 
+# Added in 2.285
+class c_nvmlHwbcEntry_t(Structure):
+    _fields_ = [
+        ('hwbcId', c_uint),
+        ('firmwareVersion', c_char * 32),
+    ]
+
+## Event structures
+class struct_c_nvmlEventSet_t(Structure):
+    pass # opaque handle
+c_nvmlEventSet_t = POINTER(struct_c_nvmlEventSet_t)
+
+nvmlEventTypeSingleBitEccError     = 0x0000000000000001
+nvmlEventTypeDoubleBitEccError     = 0x0000000000000002
+nvmlEventTypePState                = 0x0000000000000004     
+nvmlEventTypeXidCriticalError      = 0x0000000000000008     
+nvmlEventTypeNone                  = 0x0000000000000000  
+nvmlEventTypeAll                   = (
+                                        nvmlEventTypeNone |
+                                        nvmlEventTypeSingleBitEccError |
+                                        nvmlEventTypeDoubleBitEccError |
+                                        nvmlEventTypePState |
+                                        nvmlEventTypeXidCriticalError
+                                     )
+
+class c_nvmlEventData_t(Structure):
+    _fields_ = [
+        ('device', c_nvmlDevice_t),
+        ('eventType', c_ulonglong),
+        ('reserved', c_ulonglong)
+    ]
+
 ## C function wrappers ##
 def nvmlInit():
+    global nvmlLib
+    global libLoadLock
+    
+    #
+    # Load the library if it isn't loaded already
+    #
+    if (nvmlLib == None):
+        # lock to ensure only one caller loads the library
+        libLoadLock.acquire()
+        
+        try:
+            # ensure the library still isn't loaded
+            if (nvmlLib == None):
+                try:
+                    if (sys.platform[:3] == "win"):
+                        # cdecl calling convention
+                        nvmlLib = cdll.nvml
+                    else:
+                        # assume linux
+                        nvmlLib = CDLL("libnvidia-ml.so")
+                except OSError, ose:
+                    print(ose)
+                    _nvmlCheckReturn(NVML_ERROR_DRIVER_NOT_LOADED)
+                if (nvmlLib == None):
+                    print("Failed to load NVML")
+                    _nvmlCheckReturn(NVML_ERROR_DRIVER_NOT_LOADED)
+        finally:
+            # lock is always freed
+            libLoadLock.release()
+            
+    #
+    # Initialize the library
+    #
     fn = _nvmlGetFunctionPointer("nvmlInit")
     ret = fn()
     _nvmlCheckReturn(ret)
     return None
     
 def nvmlShutdown():
+    #
+    # Leave the library loaded, but shutdown the interface
+    #
     fn = _nvmlGetFunctionPointer("nvmlShutdown")
     ret = fn()
     _nvmlCheckReturn(ret)
     return None
 
+# Added in 2.285
 def nvmlErrorString(result):
-    try:
-        fn = _nvmlGetFunctionPointer("nvmlErrorString")
-        fn.restype = c_char_p # otherwise return is an int
-        ret = fn(result)
-    except NVMLError, nvmlError:
-        if (nvmlError.value == NVML_ERROR_NOT_SUPPORTED):
-            # NVML 2.0 doesn't support this function
-            # use the python version
-            ret = _compat2_0ErrorToString[result]
-        else:
-            # Unknown error
-            raise nvmlError
+    fn = _nvmlGetFunctionPointer("nvmlErrorString")
+    fn.restype = c_char_p # otherwise return is an int
+    ret = fn(result)
     return ret
-    
+
+# Added in 2.285
+def nvmlSystemGetNVMLVersion():
+    c_version = create_string_buffer(256)
+    fn = _nvmlGetFunctionPointer("nvmlSystemGetNVMLVersion")
+    ret = fn(c_version, c_uint(256))
+    _nvmlCheckReturn(ret)
+    return c_version.value
+
+# Added in 2.285
+def nvmlSystemGetProcessName(pid):
+    c_name = create_string_buffer(1024)
+    fn = _nvmlGetFunctionPointer("nvmlSystemGetProcessName")
+    ret = fn(c_uint(pid), c_name, c_uint(1024))
+    _nvmlCheckReturn(ret)
+    return c_name.value
+
 def nvmlSystemGetDriverVersion():
     c_version = create_string_buffer(100)
     fn = _nvmlGetFunctionPointer("nvmlSystemGetDriverVersion")
@@ -270,6 +388,30 @@ def nvmlSystemGetDriverVersion():
     _nvmlCheckReturn(ret)
     return c_version.value
 
+# Added in 2.285
+def nvmlSystemGetHicVersion():
+    c_count = c_uint(0)
+    hics = None
+    fn = _nvmlGetFunctionPointer("nvmlSystemGetHicVersion")
+    
+    # get the count
+    ret = fn(byref(c_count), None)
+    
+    # this should only fail with insufficient size
+    if ((ret != NVML_SUCCESS) and
+        (ret != NVML_ERROR_INSUFFICIENT_SIZE)):
+        raise NVMLError(ret)
+    
+    # if there are no hics
+    if (c_count.value == 0):
+        return []
+    
+    hic_array = c_nvmlHwbcEntry_t * c_count.value
+    hics = hic_array()
+    ret = fn(byref(c_count), hics)
+    _nvmlCheckReturn(ret)
+    return hics
+
 ## Unit get functions
 def nvmlUnitGetCount():
     c_count = c_uint()
@@ -324,7 +466,6 @@ def nvmlUnitGetFanSpeedInfo(unit):
 # added to API
 def nvmlUnitGetDeviceCount(unit):
     c_count = c_uint(0)
-
     # query the unit to determine device count
     fn = _nvmlGetFunctionPointer("nvmlUnitGetDevices")
     ret = fn(unit, byref(c_count), None)
@@ -419,7 +560,7 @@ def nvmlDeviceGetPersistenceMode(handle):
     
 def nvmlDeviceGetPciInfo(handle):
     c_info = nvmlPciInfo_t()
-    fn = _nvmlGetFunctionPointer("nvmlDeviceGetPciInfo")
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetPciInfo_v2")
     ret = fn(handle, byref(c_info))
     _nvmlCheckReturn(ret)
     return c_info
@@ -430,7 +571,15 @@ def nvmlDeviceGetClockInfo(handle, type):
     ret = fn(handle, _nvmlClockType_t(type), byref(c_clock))
     _nvmlCheckReturn(ret)
     return c_clock.value
-    
+
+# Added in 2.285
+def nvmlDeviceGetMaxClockInfo(handle, type):
+    c_clock = c_uint()
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetMaxClockInfo")
+    ret = fn(handle, _nvmlClockType_t(type), byref(c_clock))
+    _nvmlCheckReturn(ret)
+    return c_clock.value
+
 def nvmlDeviceGetFanSpeed(handle):
     c_speed = c_uint()
     fn = _nvmlGetFunctionPointer("nvmlDeviceGetFanSpeed")
@@ -444,13 +593,21 @@ def nvmlDeviceGetTemperature(handle, sensor):
     ret = fn(handle, _nvmlTemperatureSensors_t(sensor), byref(c_temp))
     _nvmlCheckReturn(ret)
     return c_temp.value
-    
+
+# DEPRECATED use nvmlDeviceGetPerformanceState
 def nvmlDeviceGetPowerState(handle):
     c_pstate = _nvmlPstates_t()
     fn = _nvmlGetFunctionPointer("nvmlDeviceGetPowerState")
     ret = fn(handle, byref(c_pstate))
     _nvmlCheckReturn(ret)
     return c_pstate.value
+    
+def nvmlDeviceGetPerformanceState(handle):
+    c_pstate = _nvmlPstates_t()
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetPerformanceState")
+    ret = fn(handle, byref(c_pstate))
+    _nvmlCheckReturn(ret)
+    return c_pstate.value
 
 def nvmlDeviceGetPowerManagementMode(handle):
     c_pcapMode = _nvmlEnableState_t()
@@ -542,6 +699,49 @@ def nvmlDeviceGetCurrentDriverModel(handle):
 def nvmlDeviceGetPendingDriverModel(handle):
     return nvmlDeviceGetDriverModel(handle)[1]
 
+# Added in 2.285
+def nvmlDeviceGetVbiosVersion(handle):
+    c_version = create_string_buffer(32)
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetVbiosVersion")
+    ret = fn(handle, c_version, c_uint(32))
+    _nvmlCheckReturn(ret)
+    return c_version.value
+
+# Added in 2.285
+def nvmlDeviceGetComputeRunningProcesses(handle):
+    # first call to get the size
+    c_count = c_uint(0)
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetComputeRunningProcesses")
+    ret = fn(handle, byref(c_count), None)
+    
+    if (ret == NVML_SUCCESS):
+        # special case, no running processes
+        return []
+    elif (ret == NVML_ERROR_INSUFFICIENT_SIZE):
+        # typical case
+        # oversize the array incase more processes are created
+        c_count.value = c_count.value * 2 + 5
+        proc_array = c_nvmlProcessInfo_t * c_count.value
+        c_procs = proc_array()
+        
+        # make the call again
+        ret = fn(handle, byref(c_count), c_procs)
+        _nvmlCheckReturn(ret)
+        
+        procs = []
+        for i in range(c_count.value):
+            # use an alternative struct for this object
+            obj = nvmlStructToFriendlyObject(c_procs[i])
+            if (obj.usedGpuMemory == NVML_VALUE_NOT_AVAILABLE_ulonglong.value):
+                # special case for WDDM on Windows, see comment above
+                obj.usedGpuMemory = None
+            procs.append(obj)
+
+        return procs
+    else:
+        # error case
+        raise NVMLError(ret)
+
 ## Set functions
 def nvmlUnitSetLedState(unit, color):
     fn = _nvmlGetFunctionPointer("nvmlUnitSetLedState")
@@ -579,5 +779,43 @@ def nvmlDeviceSetDriverModel(handle, model):
     _nvmlCheckReturn(ret)
     return None
 
-    
+# Added in 2.285
+def nvmlEventSetCreate():
+    fn = _nvmlGetFunctionPointer("nvmlEventSetCreate")
+    eventSet = c_nvmlEventSet_t()
+    ret = fn(byref(eventSet))
+    _nvmlCheckReturn(ret)
+    return eventSet
+
+# Added in 2.285
+def nvmlDeviceRegisterEvents(handle, eventTypes, eventSet):
+    fn = _nvmlGetFunctionPointer("nvmlDeviceRegisterEvents")
+    ret = fn(handle, c_ulonglong(eventTypes), eventSet)
+    _nvmlCheckReturn(ret)
+    return None
+
+# Added in 2.285
+def nvmlDeviceGetSupportedEventTypes(handle):
+    c_eventTypes = c_ulonglong()
+    fn = _nvmlGetFunctionPointer("nvmlDeviceGetSupportedEventTypes")
+    ret = fn(handle, byref(c_eventTypes))
+    _nvmlCheckReturn(ret)
+    return c_eventTypes.value
+
+# Added in 2.285
+# raises NVML_ERROR_TIMEOUT exception on timeout
+def nvmlEventSetWait(eventSet, timeoutms):
+    fn = _nvmlGetFunctionPointer("nvmlEventSetWait")
+    data = c_nvmlEventData_t()
+    ret = fn(eventSet, byref(data), c_uint(timeoutms))
+    _nvmlCheckReturn(ret)
+    return data
+
+# Added in 2.285
+def nvmlEventSetFree(eventSet):
+    fn = _nvmlGetFunctionPointer("nvmlEventSetFree")
+    ret = fn(eventSet)
+    _nvmlCheckReturn(ret)
+    return None
+
 
